@@ -3,7 +3,7 @@
 --
 luatexbase.provides_module({
   name = 'luatexja.jfont',
-  date = '2015/09/19',
+  date = '2016/04/03',
   description = 'Loader for Japanese fonts',
 })
 module('luatexja.jfont', package.seeall)
@@ -13,19 +13,14 @@ luatexja.load_module('charrange'); local ltjc = luatexja.charrange
 luatexja.load_module('rmlgbm');    local ltjr = luatexja.rmlgbm
 luatexja.load_module('direction'); local ltjd = luatexja.direction
 
+local setfield = node.direct.setfield
+local getid = node.direct.getid
+local to_direct = node.direct.todirect
 
-local Dnode = node.direct or node
-
-local setfield = (Dnode ~= node) and Dnode.setfield or function(n, i, c) n[i] = c end
-local getid = (Dnode ~= node) and Dnode.getid or function(n) return n.id end
-
-local nullfunc = function(n) return n end
-local to_direct = (Dnode ~= node) and Dnode.todirect or nullfunc
-
-local node_new = Dnode.new
-local node_free = Dnode.free
-local has_attr = Dnode.has_attribute
-local set_attr = Dnode.set_attribute
+local node_new = node.direct.new
+local node_free = node.direct.free
+local has_attr = node.direct.has_attribute
+local set_attr = node.direct.set_attribute
 local round = tex.round
 local font_getfont = font.getfont
 
@@ -34,11 +29,8 @@ local attr_curjfnt = luatexbase.attributes['ltj@curjfnt']
 local attr_curtfnt = luatexbase.attributes['ltj@curtfnt']
 local id_glyph = node.id('glyph')
 local id_kern = node.id('kern')
-local id_glue_spec = node.id('glue_spec')
-local id_glue = node.id('glue')
 local cat_lp = luatexbase.catcodetables['latex-package']
 local FROM_JFM     = luatexja.icflag_table.FROM_JFM
-local tokenlib = luatexja.token
 ------------------------------------------------------------------------
 -- LOADING JFM
 ------------------------------------------------------------------------
@@ -165,27 +157,22 @@ do
 	    ((v.align=='right') and 1 or 0.5)
 	 if type(i) == 'number' then -- char_type
 	    for k,w in pairs(v.glue) do
-	       local h = node_new(id_glue_spec)
 	       v[k] = {
-		  true, h, 
-		  ratio=w.ratio/sz, 
+		  nil,
+		  ratio=w.ratio/sz,
 		  priority=FROM_JFM + w.priority/sz,
+		  width = w[1], stretch = w[2], shrink = w[3],
 		  kanjiskip_natural = w.kanjiskip_natural and w.kanjiskip_natural/sz,
 		  kanjiskip_stretch = w.kanjiskip_stretch and w.kanjiskip_stretch/sz,
 		  kanjiskip_shrink =  w.kanjiskip_shrink  and w.kanjiskip_shrink/sz,
 	       }
-	       setfield(h, 'width', w[1])
-	       setfield(h, 'stretch', w[2])
-	       setfield(h, 'shrink', w[3])
-	       setfield(h, 'stretch_order', 0)
-	       setfield(h, 'shrink_order', 0)
 	    end
 	    for k,w in pairs(v.kern) do
 	       local g = node_new(id_kern)
 	       setfield(g, 'kern', w[1])
 	       setfield(g, 'subtype', 1)
 	       set_attr(g, attr_icflag, FROM_JFM)
-	       v[k] = {false, g, ratio=w[2]/sz}
+	       v[k] = {g, ratio=w[2]/sz}
 	    end
 	 end
 	 v.glue, v.kern = nil, nil
@@ -260,7 +247,7 @@ do
       cstemp = csname:sub( (utf8.byte(csname,1,1) == tex.escapechar) and 2 or 1, -1)
       cstemp = cstemp:sub(1, ((cstemp:sub(-1,-1)==' ') and (cstemp:len()>=2)) and -2 or -1)
       global_flag = g and '\\global' or ''
-      tex.sprint(cat_lp, '\\expandafter\\font\\csname ', 
+      tex.sprint(cat_lp, '\\expandafter\\font\\csname ',
 		 (cstemp==' ') and '\\space' or cstemp, '\\endcsname')
    end
 
@@ -277,16 +264,17 @@ do
 			    "bad JFM `" .. jfm_file_name .. "'",
 			    'The JFM file you specified is not valid JFM file.\n'..
 			       'So defining Japanese font is cancelled.')
-	 tex.sprint(cat_lp, global_flag, '\\expandafter\\let\\csname ', 
+	 tex.sprint(cat_lp, global_flag, '\\expandafter\\let\\csname ',
 		    (cstemp==' ') and '\\space' or cstemp,
 		       '\\endcsname=\\relax')
 	 return
       end
+      if not f then return end
       update_jfm_cache(j, f.size)
       local ad = identifiers[fn].parameters
       local sz = metrics[j].size_cache[f.size]
       local fmtable = { jfm = j, size = f.size, var = jfm_var,
-			with_kanjiskip = jfm_ksp, 
+			with_kanjiskip = jfm_ksp,
 			zw = sz.zw, zh = sz.zh,
 			ascent = ad.ascender,
 			descent = ad.descender,
@@ -325,6 +313,8 @@ do
    -- extract jfm_file_name and jfm_var
    -- normalize position of 'jfm=' and 'jfmvar=' keys
    local function extract_metric(name)
+      local is_braced = name:match('^{(.*)}$')
+       name= is_braced or name
       jfm_file_name = ''; jfm_var = ''; jfm_ksp = true
       local tmp, index = name:sub(1, 5), 1
       if tmp == 'file:' or tmp == 'name:' or tmp == 'psft:' then
@@ -369,16 +359,16 @@ do
       if jfm_dir == 'tate' then
 	 is_vert_enabled = (not name:match('[:;]%-vert')) and (not  name:match('[:;]%-vrt2'))
          if not name:match('vert') and not name:match('vrt2') then
-            name = name .. ';vert;vrt2'
+            name = name .. ';+vert;+vrt2'
          end
       else
 	 is_vert_enabled = nil
       end
-      return name
+      return is_braced and ('{' .. name .. '}') or name
    end
 
    -- define_font callback
-   local otfl_fdr = fonts.definers.read
+   local otfl_fdr
    local ltjr_font_callback = ltjr.font_callback
    function luatexja.font_callback(name, size, id)
       local new_name = is_def_jfont and extract_metric(name) or name
@@ -389,6 +379,7 @@ do
       return res
    end
    luatexbase.create_callback('luatexja.define_font', 'simple', function (n) return n end)
+   otfl_fdr= luatexbase.remove_from_callback('define_font', 'luaotfload.define_font')
    luatexbase.add_to_callback('define_font',luatexja.font_callback,"luatexja.font_callback", 1)
 end
 
@@ -417,9 +408,9 @@ do
       kfam_list[fam]=true
    end
    function search_kfam(fam, use_fd)
-      if kfam_list[fam] then 
+      if kfam_list[fam] then
 	 tex.sprint(cat_lp, '\\let\\ifin@\\iftrue '); return
-      elseif Nkfam_list[fam] then 
+      elseif Nkfam_list[fam] then
 	 tex.sprint(cat_lp, '\\let\\ifin@\\iffalse '); return
       elseif use_fd then
 	 for i,_ in pairs(kyenc_list) do
@@ -452,7 +443,7 @@ do
      tex.sprint(cat_lp, s)
    end
    function search_ffam_fd(fam)
-      if Nffam_list[fam] then 
+      if Nffam_list[fam] then
 	 tex.sprint(cat_lp, '\\let\\ifin@\\iffalse '); return
       else
 	 for i,_ in pairs(fenc_list) do
@@ -658,17 +649,28 @@ do
 	       for k,w in pairs(v) do
 		  if type(k)=='number' then
 		     --if w[1] then gs = gs + 1 else ke = ke + 1 end
-		     node_free(w[2])
+		     if w[1] then node_free(w[1]) end
 		  end
 	       end
 	    end
 	    n.size_cache[i]=nil
 	 end
       end
-      --print('glue spec: ', gs, 'kern: ', ke)
    end
 end
 
+------------------------------------------------------------------------
+-- VERT VARIANT TABLE
+------------------------------------------------------------------------
+local vert_form_table = {
+   [0x2013]=0xFE32, [0x2014]=0xFE31, [0x2025]=0xFE30,
+   [0xFF08]=0xFE35, [0xFF09]=0xFE36, [0xFF5B]=0xFE37, [0xFF5D]=0xFE38,
+   [0x3014]=0xFE39, [0x3015]=0xFE3A, [0x3010]=0xFE3B, [0x3011]=0xFE3C,
+   [0x300A]=0xFE3D, [0x300B]=0xFE3E, [0x3008]=0xFE3F, [0x3009]=0xFE40,
+   [0x300C]=0xFE41, [0x300D]=0xFE42, [0x300E]=0xFE43, [0x300F]=0xFE44,
+   [0xFF3B]=0xFE47, [0xFF3D]=0xFE48, [0xFF3F]=0xFE33,
+}
+
 ------------------------------------------------------------------------
 -- 追加のフォント情報
 ------------------------------------------------------------------------
@@ -693,12 +695,11 @@ do
    end
 
    local sort = table.sort
-   local function add_fl_table(dest, glyphs, unitable, asc_des, units)
-      local tg, glyphmin, glyphmax = glyphs.glyphs, 0, glyphs.glyphmax
-      for _,v in pairs(fields(glyphs)) do
-	 if v=='glyphmin' then glyphmin = glyphs.glyphmin; break end
-      end
-      for i = glyphmin, glyphmax-1 do
+   local function add_fl_table(dest, glyphs, unitable, asc_des, units, id)
+      local glyphmin, glyphmax = glyphs.glyphmin, glyphs.glyphmax
+      if glyphmax < 0 then return dest end
+      local tg = glyphs.glyphs
+      for i = glyphmin, glyphmax do
 	 local gv = tg[i]
 	 if gv then
 	    if gv.altuni then
@@ -706,17 +707,30 @@ do
 		  local bu, vsel = at.unicode, at.variant
 		  if vsel then
 		     if vsel>=0xE0100 then vsel = vsel - 0xE0100 end
-		     dest = dest or {}; dest[bu] = dest[bu] or {}
 		     local uniq_flag = true
-		     for i,_ in pairs(dest[bu]) do
-			if i==vs then uniq_flag = false; break end
-		     end
+                     if dest and dest[bu] then
+			for i,_ in pairs(dest[bu]) do
+			   if i==vsel then uniq_flag = false; break end
+		        end
+                     end
 		     if uniq_flag then
+			dest = dest or {}; dest[bu] = dest[bu] or {}
 			dest[bu][vsel] = unitable[gv.name]
 		     end
 		  end
 	       end
 	    end
+	    -- vertical form
+	    local gi = unitable[gv.name]
+	    if gi then
+	       if unitable[gv.name .. '.vert'] then
+	          dest = dest or {}; dest[gi] = dest[gi] or {};
+	          dest[gi].vform = unitable[gv.name .. '.vert']
+	       elseif id.characters[gi] and vert_form_table[gi] then
+	          dest = dest or {}; dest[gi] = dest[gi] or {};
+	          dest[gi].vform = vert_form_table[gi]
+	       end
+	    end
 	    -- vertical metric
 	    local vw, tsb, vk = glyph_vmetric(gv)
 	    local gi = unitable[gv.name]
@@ -755,19 +769,37 @@ do
    end
    prepare_fl_data = function (dest, id)
       local fl = fontloader.open(id.filename)
-      local unicodes = id.resources.unicodes
+      local ind_to_uni, unicodes = {}, {}
+      for i,v in pairs(id.characters) do
+	  ind_to_uni[v.index] = i
+      end
       if fl.glyphs then
+	 local tg, glyphmin, glyphmax = fl.glyphs, fl.glyphmin, fl.glyphmax
+         if 0 <= glyphmax then
+            for i = glyphmin, glyphmax do
+               if tg[i] and tg[i].name then unicodes[tg[i].name] = ind_to_uni[i] end
+	    end
+	 end
 	 dest = add_fl_table(dest, fl, unicodes,
-			     fl.ascent + fl.descent, fl.units_per_em)
+			     fl.ascent + fl.descent, fl.units_per_em, id)
       end
       if fl.subfonts then
          for _,v in pairs(fl.subfonts) do
-            dest = add_fl_table(dest, v, unicodes, 
-				fl.ascent + fl.descent, fl.units_per_em)
+	    local tg, glyphmin, glyphmax = v.glyphs, v.glyphmin, v.glyphmax
+            if 0 <= glyphmax then
+               for i = glyphmin, glyphmax do
+                  if tg[i] and tg[i].name then unicodes[tg[i].name] = ind_to_uni[i] end
+	       end
+	   end
+       end
+         for _,v in pairs(fl.subfonts) do
+            dest = add_fl_table(dest, v, unicodes,
+				fl.ascent + fl.descent, fl.units_per_em, id)
          end
-      end
-      fontloader.close(fl); collectgarbage("collect")
-      return dest
+     end
+     if dest then dest.unicodes = unicodes end
+     fontloader.close(fl); collectgarbage("collect")
+     return dest
    end
    -- supply vkern table
    supply_vkern_table = function(id, bname)
@@ -788,12 +820,12 @@ end
 
 --
 do
-   local cache_ver = 6
+   local cache_ver = 11
    local checksum = file.checksum
 
    local function prepare_extra_data_base(id)
-      if not id then return end
-      local bname = file.nameonly(id.filename or '')
+      if (not id) or (not id.filename) then return end
+      local bname = file.nameonly(id.filename)
       if not font_extra_basename[bname] then
 	 -- if the cache is present, read it
 	 local newsum = checksum(id.filename) -- MD5 checksum of the fontfile
@@ -820,7 +852,7 @@ do
       end
    end
    local function prepare_extra_data_font(id, res)
-      if type(res)=='table' and res.shared then
+      if type(res)=='table' and res.shared and res.filename then
 	 font_extra_info[id] = font_extra_basename[file.nameonly(res.filename)]
       end
    end
@@ -932,7 +964,7 @@ do
 
 	return tfmdata
      end,
-     'ltj.supply_tounicode', 1)  
+     'ltj.supply_tounicode', 1)
 end
 
 
@@ -940,14 +972,24 @@ end
 -- MISC
 ------------------------------------------------------------------------
 do
-   local getfont = (Dnode ~= node) and Dnode.getfont or function(n) return n.font end
-   local getchar = (Dnode ~= node) and Dnode.getchar or function(n) return n.char end
+   local getfont = node.direct.getfont
+   local getchar = node.direct.getchar
    local get_dir_count = ltjd.get_dir_count
    local is_ucs_in_japanese_char = ltjc.is_ucs_in_japanese_char_direct
    local ensure_tex_attr = ltjb.ensure_tex_attr
-   local node_write = Dnode.write
+   local node_write = node.direct.write
    local font = font
-   local ITALIC       = luatexja.icflag_table.ITALIC
+   local new_ic_kern
+   if status.luatex_version>=89 then
+       new_ic_kern = function(g)  return node_new(id_kern,3) end
+   else
+       local ITALIC       = luatexja.icflag_table.ITALIC
+       new_ic_kern = function()
+         local g = node_new(id_kern)
+         setfield(g, 'subtype', 1); set_attr(g, attr_icflag, ITALIC)
+	 return g
+       end
+   end
    -- EXT: italic correction
    function append_italic()
       local p = to_direct(tex.nest[tex.nest.ptr].tail)
@@ -957,18 +999,18 @@ do
 	    local j = font_metric_table[
 	       has_attr(p, (get_dir_count()==dir_tate) and attr_curtfnt or attr_curjfnt)
 	       ]
-	    local g = node_new(id_kern)
-	    setfield(g, 'subtype', 1); set_attr(g, attr_icflag, ITALIC)
+	    local g = new_ic_kern()
 	    setfield(g, 'kern', j.char_type[find_char_class(getchar(p), j)].italic)
 	    node_write(g); ensure_tex_attr(attr_icflag, 0)
 	 else
 	    local f = getfont(p)
 	    local h = font_getfont(f) or font.fonts[f]
 	    if h then
-	       local g = node_new(id_kern)
-	       setfield(g, 'subtype', 1); set_attr(g, attr_icflag, ITALIC)
-	       setfield(g, 'kern', h.characters[getchar(p)].italic)
-	       node_write(g); ensure_tex_attr(attr_icflag, 0)
+	       local g = new_ic_kern()
+	       if h.characters[getchar(p)] and h.characters[getchar(p)].italic then
+		  setfield(g, 'kern', h.characters[getchar(p)].italic)
+		  node_write(g); ensure_tex_attr(attr_icflag, 0)
+	       end
 	    end
 	 end
       end
@@ -976,16 +1018,3 @@ do
    end
 end
 
-------------------------------------------------------------------------
--- VERT VARIANT TABLE
-------------------------------------------------------------------------
-vert_form_table = {
-   [0x2013]=0xFE32, [0x2014]=0xFE31, [0x2025]=0xFE30,
-   [0xFF08]=0xFE35, [0xFF09]=0xFE36, [0xFF5B]=0xFE37, [0xFF5D]=0xFE38,
-   [0x3014]=0xFE39, [0x3015]=0xFE3A, [0x3010]=0xFE3B, [0x3011]=0xFE3C,
-   [0x300A]=0xFE3D, [0x300B]=0xFE3E, [0x3008]=0xFE3F, [0x3009]=0xFE40,
-   [0x300C]=0xFE41, [0x300D]=0xFE42, [0x300E]=0xFE43, [0x300F]=0xFE44,
-   [0xFF3B]=0xFE47, [0xFF3D]=0xFE48, [0xFF3F]=0xFE33,
-}
-setmetatable(vert_form_table, {__index=function(t,k) return k end});
-
