@@ -20,14 +20,16 @@ function luatexja.load_lua(fn)
    end
 end
 
--- check token library
-if newtoken then
-   luatexja.token = newtoken
-else
-   local luatex_version = status.list().luatex_version
-   if luatex_version >80 then
-      luatexja.token = token
-   end
+do
+    local setfield = node.direct.setfield
+    luatexja.setglue = node.direct.setglue or
+    function(g,w,st,sh,sto,sho)
+	setfield(g,'width', w or 0)
+	setfield(g,'stretch',st or 0)
+	setfield(g,'shrink', sh or 0)
+	setfield(g,'stretch_order',sto or 0)
+	setfield(g,'shrink_order', sho or 0)
+    end
 end
 
 --- 以下は全ファイルで共有される定数
@@ -131,6 +133,7 @@ local function print_scaled(s)
    until s<=delta
    return out
 end
+luatexja.print_scaled = print_scaled
 
 local function print_glue(d,order)
    local out=print_scaled(d)
@@ -291,10 +294,9 @@ end
 do
    local start_time_measure, stop_time_measure
       = ltjb.start_time_measure, ltjb.stop_time_measure
-   local Dnode = node.direct or node
    local nullfunc = function (n) return n end
-   local to_node = (Dnode ~= node) and Dnode.tonode or nullfunc
-   local to_direct = (Dnode ~= node) and Dnode.todirect or nullfunc
+   local to_node = node.direct.tonode
+   local to_direct = node.direct.todirect
    local ensure_tex_attr = ltjb.ensure_tex_attr
 
    -- mode = true iff main_process is called from pre_linebreak_filter
@@ -318,22 +320,20 @@ do
    end
 
    -- callbacks
-   luatexbase.add_to_callback(
+   ltjb.add_to_callback(
       'pre_linebreak_filter',
       function (head,groupcode)
 	 return main_process(head, true, tex.textdir, groupcode)
-      end,'ltj.pre_linebreak_filter',
-      luatexbase.priority_in_callback('pre_linebreak_filter',
-				      'luaotfload.node_processor') + 1)
-   luatexbase.add_to_callback(
+      end,'ltj.main',
+      luatexbase.priority_in_callback('pre_linebreak_filter', 'luaotfload.node_processor')+1)
+   ltjb.add_to_callback(
       'hpack_filter',
       function (head,groupcode,size,packtype, dir)
 	 return main_process(head, false, dir, groupcode)
-      end,'ltj.hpack_filter',
-      luatexbase.priority_in_callback('hpack_filter',
-				      'luaotfload.node_processor') + 1)
-   luatexbase.add_to_callback('pre_linebreak_filter', adjust_icflag, 'ltj.adjust_icflag', 1)
-   luatexbase.add_to_callback('hpack_filter', adjust_icflag, 'ltj.adjust_icflag', 1)
+      end,'ltj.main',
+      luatexbase.priority_in_callback('hpack_filter', 'luaotfload.node_processor')+1)
+   ltjb.add_to_callback('pre_linebreak_filter', adjust_icflag, 'ltj.adjust_icflag', 1)
+   ltjb.add_to_callback('hpack_filter', adjust_icflag, 'ltj.adjust_icflag', 1)
 
 end
 
@@ -354,7 +354,6 @@ local has_attr = node.has_attribute
 
 local id_penalty = node.id('penalty')
 local id_glyph = node.id('glyph')
-local id_glue_spec = node.id('glue_spec')
 local id_glue = node.id('glue')
 local id_kern = node.id('kern')
 local id_hlist = node.id('hlist')
@@ -381,6 +380,7 @@ local function debug_show_node_X(p,print_fn, limit)
          .. tostring(p.font)
          .. ' (' .. print_scaled(p.height) .. '+'
          .. print_scaled(p.depth) .. ')x' .. print_scaled(p.width)
+         .. ' xoff: ' .. print_scaled(p.xoffset)
       print_fn(s)
    elseif pt=='hlist' or pt=='vlist' or pt=='unset'or pt=='ins' then
       if pt=='ins' then
@@ -424,7 +424,7 @@ local function debug_show_node_X(p,print_fn, limit)
 	 .. ', dir=' .. tostring(node.has_attribute(p, attr_dir))
       print_fn(s)
    elseif pt == 'glue' then
-      s = base .. ' ' ..  print_spec(p.spec)
+      s = base .. ' ' ..  print_spec(p)
       if get_attr_icflag(p)>icflag_table.KINSOKU
          and get_attr_icflag(p)<icflag_table.KANJI_SKIP then
          s = s .. ' (from JFM: priority ' .. get_attr_icflag(p)-icflag_table.FROM_JFM .. ')'
@@ -481,12 +481,12 @@ local function debug_show_node_X(p,print_fn, limit)
             prefix, inner_depth = k, bid
          end
       else
-         s = s .. node.subtype(p.subtype)
+         s = s .. (node.subtype(p.subtype) or '')
 	 if p.subtype==1 then
 	    s = s .. ' stream=' .. p.stream
 	    print_fn(s)
 	    for i=1,#p.data do
-	       print_fn(s .. '  [' .. i .. '] = ' .. tostring(token.command_name(p.data[i])))
+	       print_fn(s .. '  [' .. i .. '] = ' .. tostring(p.data[i].csname))
 	    end
 	 else
 	    print_fn(s)
